
#include    <assert.h>
#include    <stdio.h>
#include    "libExt.h"
#include    "Torus.h"

inline  Vector3D    cross ( const Vector4D& v1, const Vector4D& v2 )
{
    return Vector3D ( v1.x, v1.y, v1.z ) ^ Vector3D ( v2.x, v2.y, v2.z );
}

///////////////////////////////////////////////////////////////////////////////////

Torus :: Torus ( float r1, float r2, int rings, int sides )
{
    float   ringDelta = 2.0 * M_PI / rings;
    float   sideDelta = 2.0 * M_PI / sides;
    float   invRings  = 1.0 / (float) rings;
    float   invSides  = 1.0 / (float) sides;
    int     i, j;

    numRings = rings;
    numSides = sides;
    vertices = new Vertex [(numSides+1)*(numRings+1)];
    faces    = new Face   [numSides*numRings*2];

                                // create vertices
    int index = 0;

    for ( i = 0; i < rings; i++ )
    {
        float theta    = i * ringDelta;
        float cosTheta = cos ( theta );
        float sinTheta = sin ( theta );

        for ( j = 0; j < sides; j++ )
        {
            float phi    = j * sideDelta;
            float cosPhi = cos ( phi );
            float sinPhi = sin ( phi );
            float dist = r2 + r1 * cosPhi;

        // (x,y,z) below is the parametric equation for a torus
        // when theta and phi both vary from 0 to pi.

                    // x = cos(theta) * (R + r * cos(phi))
                    // y = -sin(theta) * (R + r * cos(phi))
                    // z = r * sin(phi)

            vertices [index].pos.x = cosTheta * dist;
            vertices [index].pos.y = -sinTheta * dist;
            vertices [index].pos.z = r1 * sinPhi;
            vertices [index].pos.w = 1;

      // now setup texture coordinates

            vertices [index].tex.x = j * invSides;
            vertices [index].tex.y = i * invRings;

      // normalize the partial derivative of (x,y,z) with respect to theta.
      // T = normalize([dx/dtheta,dy/dtheta,dz/dtheta])

            float   dxdtheta = -sinTheta * dist;
            float   dydtheta = -cosTheta * dist;
            float   d        = 1 / sqrt ( dxdtheta*dxdtheta + dydtheta*dydtheta );

            vertices [index].t.x = dxdtheta * d;
            vertices [index].t.y = dydtheta * d;
            vertices [index].t.z = 0;

        // compute unit normal

            vertices [index].n.x =  cosTheta * cosPhi;
            vertices [index].n.y = -sinTheta * cosPhi;
            vertices [index].n.z = sinPhi;

        // B = N cross T

            vertices [index].b = vertices [index].n ^ vertices [index].t;

        // Note:  by the construction above, [T, B, N] form an orthonormal space.

            index++;
        }
    }

    numVertices = index;

                                            // Create faces
    index = 0;

    for ( i = 0; i < rings; i++ )
        for ( j = 0; j < sides; j++, index += 2 )
        {
            faces [index].index [0] = indexFor ( i, j );
            faces [index].index [2] = indexFor ( i + 1, j );
            faces [index].index [1] = indexFor ( i + 1, j + 1 );
            faces [index].n         = cross ( vertices [faces[index].index [1]].pos - vertices [faces[index].index [0]].pos,
                                               vertices [faces[index].index [2]].pos - vertices [faces[index].index [0]].pos );

            faces [index+1].index [0] = indexFor ( i, j );
            faces [index+1].index [2] = indexFor ( i + 1, j + 1 );
            faces [index+1].index [1] = indexFor ( i, j + 1 );
            faces [index+1].n         = cross ( vertices [faces[index+1].index [1]].pos - vertices [faces[index+1].index [0]].pos,
                                                 vertices [faces[index+1].index [2]].pos - vertices [faces[index+1].index [0]].pos );

			assert ( (faces [index].n & vertices [faces[index].index [0]].n) > 0 );
			assert ( (faces [index+1].n & vertices [faces[index+1].index [0]].n) > 0 );
        }

    numFaces = index;

                                            // create edges
    edges    = new Edge [3*numFaces];       // use max size
    numEdges = 0;

    for ( i = 0; i < numFaces; i++ )
    {
        addEdge ( faces [i].index [0], faces [i].index [1], i );
        addEdge ( faces [i].index [1], faces [i].index [2], i );
        addEdge ( faces [i].index [2], faces [i].index [0], i );
    }
}

void    Torus :: addEdge ( int a, int b, int face )
{
    int     a1  = a;
    int     b1  = b;
    bool    rev = false;
    
    if ( a > b )
    {
	a1  = b;
	b1  = a;
	rev = true;
    }

                                            // check whether we already have this face
    for ( int i = 0; i < numEdges; i++ )
        if ( edges [i].a == a1 && edges [i].b == b1 )
        {
            assert ( edges [i].f2 == -1 );

            edges [i].f2   = face;
            edges [i].rev2 = rev;

            return;
        }

                                            // insert new adge
    edges [numEdges].a    = a1;
    edges [numEdges].b    = b1;
    edges [numEdges].f1   = face;
    edges [numEdges].rev1 = rev;
    edges [numEdges].f2   = -1;

    numEdges++;
}

void    Torus :: draw ()
{
    glEnable   ( GL_CULL_FACE );
    glCullFace ( GL_BACK );

    glBegin ( GL_TRIANGLES );

    for ( int i = 0; i < numFaces; i++ )
        for ( int j = 0; j < 3; j++ )
        {
            Vertex& v = vertices [faces [i].index [j]];

                                            // send all data for vertex program
            glMultiTexCoord2fv ( GL_TEXTURE0_ARB, v.tex );
            glMultiTexCoord3fv ( GL_TEXTURE1_ARB, v.t   );
            glMultiTexCoord3fv ( GL_TEXTURE2_ARB, v.b   );
            glNormal3fv        ( v.n                    );
            glVertex4fv        ( v.pos                  );
        }

    glEnd   ();
}

void    Torus :: drawUnlit ()
{
    glEnable   ( GL_CULL_FACE );
    glCullFace ( GL_BACK      );

    glBegin ( GL_TRIANGLES );

    for ( int i = 0; i < numFaces; i++ )
        for ( int j = 0; j < 3; j++ )
        {
            glVertex4fv        ( vertices [faces [i].index [j]].pos );
            glMultiTexCoord2fv ( GL_TEXTURE0_ARB, vertices [faces [i].index [j]].tex );
        }

	glDisable ( GL_CULL_FACE );
    glEnd   ();
}

//////////////////////////// end of torus methods //////////////////////////////////
